import 'package:http/http.dart';

import 'package:requests/requests.dart';
import 'package:requests/src/response.dart';
import 'package:requests/src/common.dart';
import 'package:requests/src/cookie.dart';
import 'package:test/test.dart';

void _validateResponse(Response r) {
  expect(r.statusCode, isA<int>());
  expect(r.hasError, isA<bool>());
  expect(r.success, isA<bool>());
}

void main() {
  group('A group of tests', () {
    final String PLACEHOLDER_PROVIDER = 'https://reqres.in';

    test('plain http get', () async {
      var r = await Requests.get('https://google.com');
      r.raiseForStatus();
      dynamic body = r.content();
      expect(body, isNotNull);
      _validateResponse(r);
    });

    test('plain http get with query parameters', () async {
      var r = await Requests.get('https://google.com',
          queryParameters: {'id': 1, 'name': null});
      r.raiseForStatus();
      dynamic body = r.content();
      expect(body, isNotNull);
      expect(r.url.toString(), contains('?id=1'));
      _validateResponse(r);
    });

    test('plain http get with port 80', () async {
      var r = await Requests.get('http://google.com', port: 80);
      r.raiseForStatus();
      _validateResponse(r);
      dynamic body = r.content();
      expect(body, isNotNull);
    });

    test('plain http get with port 8080', () async {
      var r =
          await Requests.get('http://portquiz.net:8080/', timeoutSeconds: 30);
      r.raiseForStatus();
    });

    test('json http get list of objects', () async {
      var r = await Requests.get('$PLACEHOLDER_PROVIDER/api/users');
      r.raiseForStatus();
      dynamic body = r.json();
      expect(body, isNotNull);
      expect(body['data'], isList);
      _validateResponse(r);
    });

    test('FormURLEncoded http post', () async {
      var r = await Requests.post('$PLACEHOLDER_PROVIDER/api/users',
          body: {
            'userId': 10,
            'id': 91,
            'title': 'aut amet sed',
            'body':
                'libero voluptate eveniet aperiam sed\nsunt placeat suscipit molestias\nsimilique fugit nam natus\nexpedita consequatur consequatur dolores quia eos et placeat',
          },
          bodyEncoding: RequestBodyEncoding.FormURLEncoded);
      r.raiseForStatus();
      dynamic body = r.json();
      expect(body, isNotNull);
      _validateResponse(r);
    });

    test('http post a list of object', () async {
      var r = await Requests.post('$PLACEHOLDER_PROVIDER/api/users', json: [
        {
          'userId': 10,
          'id': 91,
          'title': 'aut amet sed',
          'body':
              'libero voluptate eveniet aperiam sed\nsunt placeat suscipit molestias\nsimilique fugit nam natus\nexpedita consequatur consequatur dolores quia eos et placeat',
        }
      ]);
      r.raiseForStatus();
      dynamic body = r.json();
      expect(body, isNotNull);
      _validateResponse(r);
    });

    test('json http delete with request body', () async {
      var r = await Requests.delete(
        '$PLACEHOLDER_PROVIDER/api/users/10',
        json: {'something': 'something'},
      );
      r.raiseForStatus();
      _validateResponse(r);
    });

    test('json http post', () async {
      var r = await Requests.post('$PLACEHOLDER_PROVIDER/api/users', json: {
        'userId': 10,
        'id': 91,
        'title': 'aut amet sed',
        'body':
            'libero voluptate eveniet aperiam sed\nsunt placeat suscipit molestias\nsimilique fugit nam natus\nexpedita consequatur consequatur dolores quia eos et placeat',
      });
      r.raiseForStatus();
      dynamic body = r.json();
      expect(body, isNotNull);
      _validateResponse(r);
    });

    test('json http delete', () async {
      var r = await Requests.delete('$PLACEHOLDER_PROVIDER/api/users/10');
      r.raiseForStatus();
      _validateResponse(r);
    });

    test('json http post as a form and as a JSON', () async {
      var r = await Requests.post('$PLACEHOLDER_PROVIDER/api/users', json: {
        'userId': 10,
        'id': 91,
        'title': 'aut amet sed',
        'body':
            'libero voluptate eveniet aperiam sed\nsunt placeat suscipit molestias\nsimilique fugit nam natus\nexpedita consequatur consequatur dolores quia eos et placeat',
      });
      r.raiseForStatus();
      dynamic body = r.json();
      expect(body['userId'], 10);
      _validateResponse(r);
    });

    test('json http get object', () async {
      var r = await Requests.get('$PLACEHOLDER_PROVIDER/api/users/2');
      r.raiseForStatus();
      dynamic body = r.json();
      expect(body, isNotNull);
      expect(body, isMap);
      _validateResponse(r);
    });

    test('remove cookies', () async {
      String url = '$PLACEHOLDER_PROVIDER/api/users/1';
      String hostname = Requests.getHostname(url);
      expect('reqres.in', hostname);
      await Requests.clearStoredCookies(hostname);
      var cookies = CookieJar.parseCookiesString("session=bla");
      await Requests.setStoredCookies(hostname, cookies);
      cookies = await Requests.getStoredCookies(hostname);
      expect(cookies.keys.length, 1);
      await Requests.clearStoredCookies(hostname);
      cookies = await Requests.getStoredCookies(hostname);
      expect(cookies.keys.length, 0);
    });

    test('add cookies', () async {
      String hostname = 'example';
      await Requests.addCookie(hostname, 'name', 'value');
      var cookies = await Requests.getStoredCookies(hostname);
      expect(cookies.keys.length, 1);
      await Requests.addCookie(hostname, 'name', 'value');
      cookies = await Requests.getStoredCookies(hostname);
      expect(cookies.keys.length, 1);
      await Requests.addCookie(hostname, 'another name', 'value');
      cookies = await Requests.getStoredCookies(hostname);
      expect(cookies.keys.length, 2);
      await Requests.clearStoredCookies(hostname);
      cookies = await Requests.getStoredCookies(hostname);
      expect(cookies.keys.length, 0);
    });

    test('response as Response object', () async {
      var r = await Requests.post('$PLACEHOLDER_PROVIDER/api/users',
          body: {'name': 'morpheus'});
      r.raiseForStatus();
      var content = r.content();
      var json = r.json();
      expect(content, isNotNull);
      expect(json, isNotNull);
      _validateResponse(r);
    });

    test('throw error', () async {
      try {
        var r = await Requests.get('$PLACEHOLDER_PROVIDER/api/unknown/23');
        r.raiseForStatus();
      } on HTTPException catch (e) {
        expect(e.response, isA<Response>());
        expect(e.response.success, false);
        return;
      }
      throw Exception('Expected request error');
    });

    test('throw if both json and body used', () async {
      try {
        await Requests.post('$PLACEHOLDER_PROVIDER/api/unknown/23',
            body: {}, json: {});
      } on ArgumentError catch (_) {
        return;
      }
      throw Exception('Expected request error');
    });

    test('ssl should fail due to expired certificate', () async {
      try {
        var r = await Requests.get('https://expired.badssl.com/');
        r.raiseForStatus();
      } on Exception catch (_) {
        return;
      }

      throw Exception('Expected ssl error');
    });

    test('ssl allow invalid', () async {
      var r = await Requests.get('https://expired.badssl.com/', verify: false);
      r.raiseForStatus();
    });

    test('multiple Set-Cookie response header', () async {
      var r = await Requests.get("http://samesitetest.com/cookies/set");
      var cookies = await Requests.extractResponseCookies(r.headers);

      expect(
        cookies["StrictCookie"]!.output(),
        "Set-Cookie: StrictCookie=Cookie set with SameSite=Strict; path=/; httponly; samesite=strict",
      );
      expect(
        cookies["LaxCookie"]!.output(),
        "Set-Cookie: LaxCookie=Cookie set with SameSite=Lax; path=/; httponly; samesite=lax",
      );
      expect(
        cookies["SecureNoneCookie"]!.output(),
        "Set-Cookie: SecureNoneCookie=Cookie set with SameSite=None and Secure; path=/; secure; httponly; samesite=none",
      );
      expect(
        cookies["NoneCookie"]!.output(),
        "Set-Cookie: NoneCookie=Cookie set with SameSite=None; path=/; httponly; samesite=none",
      );
      expect(
        cookies["DefaultCookie"]!.output(),
        "Set-Cookie: DefaultCookie=Cookie set without a SameSite attribute; path=/; httponly",
      );
    });

    test('cookie parsing', () async {
      var headers = Map<String, String>();
      var cookiesString = """
        session=mySecret; path=/myPath; expires=Xxx, x-x-x x:x:x XXX,
        data=1=2=3=4; _ga=GA1.4..1563550573; ; ; ; textsize=NaN; tp_state=true; _ga=GA1.3..1563550573,
        __browsiUID=03b1cb22-d18d-&{"bt":"Browser","os":"Windows","osv":"10.0","m":"Desktop|Emulator","v":"Unknown","b":"Chrome","p":2},
        _cb_ls=1; _cb=CaBNIWCf-db-3i9ro; _chartbeat2=..414141414.1..1; AMUUID=%; _fbp=fb.2..,
        adblockerfound=true 
      """;
      headers['set-cookie'] = cookiesString;
      var cookies = await Requests.extractResponseCookies(headers);

      expect(
        cookies["session"]!.output(),
        "Set-Cookie: session=mySecret; path=/myPath; expires=Xxx, x-x-x x:x:x XXX",
      );

      expect(
        cookies['data']!.output(),
        "Set-Cookie: data=1=2=3=4",
      );

      expect(
        cookies['__browsiUID']!.output(),
        'Set-Cookie: __browsiUID=03b1cb22-d18d-&{"bt":"Browser","os":"Windows","osv":"10.0","m":"Desktop|Emulator","v":"Unknown","b":"Chrome","p":2}',
      );

      expect(
        cookies['_cb_ls']!.output(),
        "Set-Cookie: _cb_ls=1",
      );

      expect(
        cookies['adblockerfound']!.output(),
        "Set-Cookie: adblockerfound=true",
      );
    });

    test('from json', () async {
      expect(Common.fromJson('{"a":1}'), {"a": 1});
      expect(Common.fromJson(null), null);
    });
  });
}
